%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Dynamic Syntax
%
% lambda calculus utilities
%
% Matthew Purver, 2002
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- use_module( library( lists ), [member/2, append/3] ).


% beta_reduce( ?Function, ?Argument, ?Result )
% copy, apply and check unbound vars
beta_reduce( (AT>PT):F, AT:A, PT:P ) :-
	nonvar( F ),
	copy_term( F, A^P ),
	unbound_vars( F, [], V ),
	unbound_vars( A^P, [], V ).


% unbound_vars( +LambdaTerm, +BoundVarList, ?UnboundVarList )
% deal with vars first - if bound, ignore ...
unbound_vars( V, Bound, [] ) :-
	member( B, Bound ),
	V == B,
	!.
% ... otherwise list as unbound
unbound_vars( V, _Bound, [V] ) :-
	var( V ),
	!.
% base case
unbound_vars( [], _Bound, [] ) :-
	!.
% recurse through arguments, collecting unbound list
unbound_vars( [H | T], Bound, Unbound ) :-
	!,
	unbound_vars( H, Bound, U1 ),
	unbound_vars( T, Bound, U2 ),
	append( U1, U2, Unbound ).
% start here: collect bound vars
unbound_vars( A^B, Bound, Unbound ) :-
	!,
	unbound_vars( B, [A | Bound], Unbound ).
% once all bound vars collected, examine arguments
unbound_vars( Expr, Bound, Unbound ) :-
	Expr =.. [_F | Args],
	unbound_vars( Args, Bound, Unbound ).
