%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Dynamic Syntax
%
% lexical action application
%
% Matthew Purver, 2002
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- use_module( library( lists ), [member/2, substitute/4] ).


% exec/3 - exec( +Tree1, +Action, ?Tree2 )
% Executes Action on Tree1, resulting in Tree2

% recurse through a list of actions
exec( Tree, [], Tree ).
exec( Tree1, [H | T], Tree3 ) :-
	exec( Tree1, H, Tree2 ),
	exec( Tree2, T, Tree3 ).

% if-then-else
exec( Tree1, ( if If, then Then, else Else ), Tree2 ) :-
	dbg( 'exec: if_then_else' ), dbgnl,
	(
	  check( Tree1, If )
	->
	  exec( Tree1, Then, Tree2 )
	;
	  exec( Tree1, Else, Tree2 )
	).

% make node
exec( tree( Nodes, Pointer ), make( Dir ), tree( [node( Node, [] ) | Nodes], Pointer )  ) :-
	dbg( 'exec: make ' ),
	find_node( Pointer, Dir, Node ),
	dbg( Node ), dbgnl.

% go to node
exec( tree( Nodes, Pointer1 ), go( Dir ), tree( Nodes, Pointer2 )  ) :-
	dbg( 'exec: go to ' ),
	member( node( Pointer2, _ ), Nodes ),
	find_node( Pointer1, Dir, Pointer2 ),
	dbg( Pointer2 ), dbgnl.

% put label
exec( tree( Nodes1, Pointer ), put( Label ), tree( Nodes2, Pointer ) ) :-
	dbg( 'exec: put ' ), dbg( Label ), dbg( ' at ' ), dbg( Pointer ), dbgnl,
	member( node( Pointer, Labels ), Nodes1 ),
	substitute( node( Pointer, Labels ), Nodes1,
		    node( Pointer, [Label | Labels] ), Nodes2 ).

% abort
exec( _, abort, _ ) :-
	dbg( 'exec: abort' ), dbgnl, dbgnl,
	fail.


% check/2 - check( +Tree, +Expr )
% Succeeds if Expr holds of Tree

% recurse through a list of expressions
check( _Tree, [] ).
check( Tree, [H | T] ) :-
	check( Tree, H ),
	check( Tree, T ).

% check local label
check( tree( Nodes, Pointer ), Label ) :-
	dbg( 'chck: ' ), dbg( Label ), dbg( ' at ' ), dbg( Pointer ), dbgnl,
	member( node( Pointer, Labels ), Nodes ),
	member( Label, Labels ).

% check local feature allowing for feature hierarchy
check( tree( Nodes, Pointer ), +Feat ) :-
	dbg( 'chck: +' ), dbg( Feat ), dbg( ' at ' ), dbg( Pointer ), dbgnl,
	member( node( Pointer, Labels ), Nodes ),
	member( +Feat2, Labels ),
	isa( Feat2, Feat ).

% check remote label
check( tree( Nodes, Pointer ), ( Dir, Label ) ) :-
	dbg( 'chck: ' ), dbg( Label ), dbg( ' at <' ), dbg( Dir ), dbg( '> ' ), dbgnl,
	member( node( Node, _ ), Nodes ), % to prevent infinite finding ...
	find_node( Pointer, Dir, Node ),
	check( tree( Nodes, Node ), Label ),
	dbg( 'chck: found at ' ), dbg( Node ), dbgnl.
